/*
 * Copyright (C) 1998-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fade.c A volume fader and alarm clock for OSS. */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <stdlib.h> /* EXIT_SUCCESS */
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <sys/soundcard.h>

#include "fade.cmdline.h"
#include "para.h"
#include "fd.h"
#include "string.h"
#include "error.h"

INIT_FADE_ERRLISTS;
static struct fade_args_info conf;

__printf_2_3 void para_log(__a_unused int ll, const char *fmt, ...)
{
	va_list argp;
	time_t t1;
	struct tm *tm;

	time(&t1);
	tm = localtime(&t1);
	printf("%d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
	va_start(argp, fmt);
	vprintf(fmt, argp);
	va_end(argp);
}

/*
 * Open the mixer device.
 */
static int open_mixer(void)
{
	return para_open(conf.mixer_device_arg, O_RDWR, 42);
}

/*
 * Get channel volume via mixer_fd.
 */
static int get_mixer_channel(int mixer_fd)
{
	int volume;

	if (ioctl(mixer_fd, MIXER_READ(conf.mixer_channel_arg), &volume) < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	/* take the mean value of left and right volume */
	return (volume % 256 + (volume >> 8)) / 2;
}

/*
 * Open mixer, get volume and close mixer.
 */
static int open_and_get_mixer_channel(void)
{
	int mixer_fd;
	int volume;

	mixer_fd = open_mixer();
	if (mixer_fd < 0)
		return mixer_fd;
	volume = get_mixer_channel(mixer_fd);
	close(mixer_fd);
	return volume;
}

/*
 * Set channel volume via mixer_fd.
 */
static int set_mixer_channel(int mixer_fd, int volume)
{
	int tmp = (volume << 8) + volume;

	if (ioctl(mixer_fd, MIXER_WRITE(conf.mixer_channel_arg), &tmp) < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/*
 * Open mixer, set volume and close mixer.
 */
static int open_and_set_mixer_channel(int volume)
{
	int mixer_fd, ret = open_mixer();

	if (ret < 0)
		return ret;
	mixer_fd = ret;
	ret = set_mixer_channel(mixer_fd, volume);
	close(mixer_fd);
	return ret;
}

static void fixup_mixer_channel_arg(void)
{
	int val = SOUND_MIXER_VOLUME; /* STFU, gcc */

	switch (conf.mixer_channel_arg) {
		case mixer_channel_arg_volume: val = SOUND_MIXER_VOLUME; break;
		case mixer_channel_arg_bass: val = SOUND_MIXER_BASS; break;
		case mixer_channel_arg_treble: val = SOUND_MIXER_TREBLE; break;
		case mixer_channel_arg_synth: val = SOUND_MIXER_SYNTH; break;
		case mixer_channel_arg_pcm: val = SOUND_MIXER_PCM; break;
		case mixer_channel_arg_speaker: val = SOUND_MIXER_SPEAKER; break;
		case mixer_channel_arg_line: val = SOUND_MIXER_LINE; break;
		case mixer_channel_arg_mic: val = SOUND_MIXER_MIC; break;
		case mixer_channel_arg_cd: val = SOUND_MIXER_CD; break;
		case mixer_channel_arg_imix: val = SOUND_MIXER_IMIX; break;
		case mixer_channel_arg_altpcm: val = SOUND_MIXER_ALTPCM; break;
		case mixer_channel_arg_reclev: val = SOUND_MIXER_RECLEV; break;
		case mixer_channel_arg_igain: val = SOUND_MIXER_IGAIN; break;
		case mixer_channel_arg_ogain: val = SOUND_MIXER_OGAIN; break;
	}
	conf.mixer_channel_arg = val;
}

/*
 * Open mixer, get volume, fade to new_vol in secs seconds and
 * close mixer.
 */
static int fade(int new_vol, int fade_time)
{
	int vol, mixer_fd, diff, incr, ret;
	unsigned secs;
	struct timespec ts;
	unsigned long long tmp, tmp2; /* Careful with that axe, Eugene! */

	if (fade_time <= 0)
		return 1;
	secs = fade_time;
	PARA_NOTICE_LOG("fading to %d in %d seconds\n", new_vol, secs);
	ret = open_mixer();
	if (ret < 0)
		return ret;
	mixer_fd = ret;
	ret = get_mixer_channel(mixer_fd);
	if (ret < 0)
		goto out;
	vol = ret;
	diff = new_vol - vol;
	if (!diff) {
		sleep(secs);
		goto out;
	}
	incr = diff > 0? 1: -1;
	diff = diff > 0? diff: -diff;
	tmp = secs * 1000 / diff;
	tmp2 = tmp % 1000;
	while ((new_vol - vol) * incr > 0) {
		ts.tv_nsec = tmp2 * 1000000; /* really nec ?*/
		ts.tv_sec = tmp / 1000; /* really nec ?*/
		//printf("ts.tv_sec: %i\n", ts.tv_nsec);
		vol += incr;
		ret = set_mixer_channel(mixer_fd, vol);
		if (ret < 0)
			goto out;
		//printf("vol = %i\n", vol);
		nanosleep(&ts, NULL);
	}
out:
	close(mixer_fd);
	return ret;
}

static void client_cmd(const char *cmd)
{
	int ret, fds[3] = {0, 0, 0};
	pid_t pid;
	char *cmdline = make_message(BINDIR "/para_client %s", cmd);

	PARA_INFO_LOG("%s\n", cmdline);
	ret = para_exec_cmdline_pid(&pid, cmdline, fds);
	free(cmdline);
	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	do
		ret = wait(NULL);
	while (ret != -1 && errno != ECHILD);
}

static void change_afs_mode_and_play(char *afs_mode)
{
	char *cmd;

	client_cmd("stop");
	if (!afs_mode)
		return;
	cmd = make_message("select %s\n", afs_mode);
	client_cmd(cmd);
	free(cmd);
	client_cmd("play");
}

/*
 * sleep
 */
static int sweet_dreams(void)
{
	time_t t1, wake_time_epoch;
	unsigned int delay;
	struct tm *tm;
	int ret, min = conf.wake_min_arg;
	char *fo_mood = conf.fo_mood_arg;
	char *fi_mood = conf.fi_mood_arg;
	char *sleep_mood = conf.sleep_mood_arg;
	int fit = conf.fi_time_arg;
	int fot = conf.fo_time_arg;
	int fiv = conf.fi_vol_arg;
	int fov = conf.fo_vol_arg;
	int iv = conf.ivol_arg;

	/* calculate wake time */
	time(&t1);
	if (conf.wake_hour_given) {
		int hour = conf.wake_hour_arg;
		tm = localtime(&t1);
		if (tm->tm_hour > hour || (tm->tm_hour == hour && tm->tm_min> min)) {
			t1 += 86400; /* wake time is tomorrow */
			tm = localtime(&t1);
		}
		tm->tm_hour = hour;
		tm->tm_min = min;
		tm->tm_sec = 0;
	} else {
		t1 += 9 * 60 * 60; /* nine hours from now */
		PARA_INFO_LOG("default wake time: %lu\n", (long unsigned)t1);
		tm = localtime(&t1);
	}
	wake_time_epoch = mktime(tm);
	PARA_INFO_LOG("waketime: %s", asctime(tm));
	client_cmd("stop");
	sleep(1);
	if (fot) {
		PARA_INFO_LOG("initial volume: %d\n", iv);
		ret = open_and_set_mixer_channel(iv);
		if (ret < 0)
			return ret;
		change_afs_mode_and_play(fo_mood);
		ret = fade(fov, fot);
		if (ret < 0)
			return ret;
	} else {
		ret = open_and_set_mixer_channel(fov);
		if (ret < 0)
			return ret;
	}
	if (conf.sleep_mood_given)
		change_afs_mode_and_play(sleep_mood);
	else
		client_cmd("stop");
	if (!fit)
		return 1;
	for (;;) {
		time(&t1);
		if (wake_time_epoch <= t1 + fit)
			break;
		delay = wake_time_epoch - t1 - fit;
		PARA_INFO_LOG("sleeping %u seconds (%u:%02u)\n",
			delay, delay / 3600,
			(delay % 3600) / 60);
		sleep(delay);
	}
	change_afs_mode_and_play(fi_mood);
	ret = fade(fiv, fit);
	PARA_INFO_LOG("fade complete, returning\n");
	return ret;
}

static int snooze(void)
{
	int ret;
	unsigned sleep_time;

	if (conf.so_time_arg <= 0)
		return 1;
	sleep_time = conf.so_time_arg;
	if (open_and_get_mixer_channel() < conf.so_vol_arg)
		ret = open_and_set_mixer_channel(conf.so_vol_arg);
	else
		ret = fade(conf.so_vol_arg, conf.so_time_arg);
	if (ret < 0)
		return ret;
	client_cmd("pause");
	PARA_NOTICE_LOG("%d seconds snooze time...\n", conf.snooze_time_arg);
	sleep(sleep_time);
	client_cmd("play");
	return fade(conf.si_vol_arg, conf.si_time_arg);
}

static int configfile_exists(void)
{
	static char *config_file;

	if (!conf.config_file_given) {
		char *home = para_homedir();
		free(config_file);
		config_file = make_message("%s/.paraslash/fade.conf", home);
		free(home);
		conf.config_file_arg = config_file;
	}
	return file_exists(conf.config_file_arg);
}

int main(int argc, char *argv[])
{
	int ret;

	if (fade_cmdline_parser(argc, argv, &conf))
		exit(EXIT_FAILURE);
	HANDLE_VERSION_FLAG("fade", conf);
	ret = configfile_exists();
	if (!ret && conf.config_file_given) {
		PARA_EMERG_LOG("can not read config file %s\n",
			conf.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (ret) {
		struct fade_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 1
		};
		fade_cmdline_parser_config_file(conf.config_file_arg,
			&conf, &params);
	}
	fixup_mixer_channel_arg();
	switch (conf.mode_arg) {
	case mode_arg_sleep:
		ret = sweet_dreams();
		break;
	case mode_arg_fade:
		ret = fade(conf.fade_vol_arg, conf.fade_time_arg);
		break;
	case mode_arg_snooze:
		ret = snooze();
		break;
	}
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
